// Package kcp is a Reliable-UDP library for Go.
//
// This library intends to provide a smooth, resilient, ordered,
// error-checked and anonymous delivery of streams over UDP packets.
//
// The interfaces of this package aim to be compatible with
// net.Conn in the standard library, but offer powerful features for advanced users.
package kcp
